Attribute VB_Name = "modLeggeBMP"
'===========================================================
' Descrizione.....: Routines per leggere informazioni e aree
'                   di pixels da files BMP.
' Nome dei Files..: LeggeBMP.bas, StruttureBMP.bas
' Data............: 13/8/2000
' Versione........: 1.0 a 32 bits
' Sistema.........: Visual Basic 6.0
'                   sotto Windows NT 4.0 (SP 3).
' Scritto da......: F. Languasco 
' E-Mail..........: MC7061@mclink.it
'===========================================================
'
'   Comprende:
'    LeggeFileInfoBMP(FileBMP$, bmfHeader As BITMAPFILEHEADER_Type _
'                   , bi As BITMAPINFO_Type, TCol&())
'     FileBMP$:     Nome del file da cui si vogliono
'                   leggere le informazioni.
'     bmfHeader:    Ritornano le strutture BMP con
'     bi:           le informazioni.
'     TCol():       Ritorna la tabella dei colori in formato RGB.
'
'    LeggeAreaBMP(MP&(), X1&, Y1&, X2&, Y2&, Termina As Boolean)
'     MP():         Ritorna, in formato RGB, la matrice dei colori
'                   dei pixels letti.
'     X1, Y1:       Coordinate definenti l' area
'     X2, Y2:       di pixels da leggere dal file.
'     Termina:      Flag di interruzione della lettura.
Option Explicit
'
Private bmfHeader As BITMAPFILEHEADER_Type
Private bi As BITMAPINFO_Type
Private TCol&()      ' Tabella dei Colori [RGB].
'
Private OffSet&     ' Posizione di inizio Immagine nel File BMP  [Bytes].
Private BMPWidth&   ' Larghezza dell' Immagine BMP [Pixels].
Private BMPHeight&  ' Altezza dell' Immagine BMP [Pixels].
Private XBMP&       ' Larghezza dell' Immagine BMP [Bytes].
'
Private BxP&        ' Profondita' di colore dell' Immagine [BitsPerPixel].
'
Private FB%         ' # del file BMP correntemente aperto.

Private Function ScanAlign(ByVal Valore&) As Long
'
'   Ritorna un multiplo di 4 uguale o immediatamente
'   superiore a Valore&.
'
    ScanAlign = (Valore + 3&) And &HFFFFFFFC
'
'
'
End Function


Public Function LeggeFileInfoBMP(ByVal FileBMP$ _
    , bmfHeader_I As BITMAPFILEHEADER_Type _
    , bi_I As BITMAPINFO_Type, TCol_I&()) As Boolean
'
'   Legge le informazioni ed i colori del file BMP:
'
    Dim C&, NCol&
'
    On Error GoTo LeggeFileInfoBMP_ERR
'
    If FB > 0 Then Close (FB)
    FB = FreeFile
    Open FileBMP$ For Binary As #FB
'
    Get #FB, , bmfHeader
    Get #FB, , bi.bmiHeader
'
    If bi.bmiHeader.biBitCount <= 8 Then
        NCol = 2 ^ bi.bmiHeader.biBitCount
        For C = 0 To NCol - 1
            Get #FB, , bi.bmiColors(C)
        Next C
        PreparaTavolaColori NCol
    End If
'
    OffSet = bmfHeader.bfOffBits + 1                ' [Bytes].
    BxP = bi.bmiHeader.biBitCount                   ' [BitsPerPixel].
    BMPWidth = bi.bmiHeader.biWidth                 ' [Pixels].
    BMPHeight = bi.bmiHeader.biHeight               ' [Pixels].
    XBMP = ScanAlign(BMPWidth * BxP / 8)   ' [Bytes].
'
    bmfHeader_I = bmfHeader
    bi_I = bi
    TCol_I() = TCol()
'
'
LeggeFileInfoBMP_ERR:
    LeggeFileInfoBMP = (Err = 0) And (bmfHeader.bfType = "BM")
'
'
'
End Function

Private Sub PreparaTavolaColori(ByVal NCol&)
'
'   Prepara la tavola dei colori in formato RGB:
'
    Dim C&
'
    ReDim TCol(0 To NCol - 1)
'
    For C = 0 To NCol - 1
        TCol(C) = RGB(bi.bmiColors(C).rgbRed _
                    , bi.bmiColors(C).rgbGreen _
                    , bi.bmiColors(C).rgbBlue)
    Next C
'
'
'
End Sub

Public Function LeggeAreaBMP(MP&() _
    , Optional ByVal X1& = 0, Optional ByVal Y1& = 0 _
    , Optional ByVal X2& = 1, Optional ByVal Y2& = 1 _
    , Optional ByRef Termina As Boolean = False) As Boolean
'
'   Ritorna, nella matrice MP(Xn, Ym), i valori RGB dei pixels
'   compresi nell' area X1 -> X2 e Y1 -> Y2 (estremi inclusi).
'   Il punto (X1, Y1) e' l' angolo superiore sinistro dell' area
'   che si vuole leggere e (X2, Y2) l' angolo inferiore destro.
'   Il punto MP(0, 0) corrisponde all' angolo superiore sinistro.
'   Nell' applicazione chiamante impostare, eventualmente,
'   Termina = True per interrompere questa routine.
'   I valori vengono letti direttamente dal file e quindi l' uso
'   di questa routine deve SEMPRE essere preceduto da una chiamata
'   a LeggeFileInfoBMP(NomeDelFileBMP$) per aprire il file e per
'   ottenere le informazioni necessarie.
'
    Dim C&, C1&, C2&, Cl&, Cl1&, Col&()
    Dim XIP&, InX0&, InX&, BnX&
    Dim YIP&, InY&, BnY&
    Dim Pi3B&, InB0&, InB&
'
    ReDim MP(0 To X2 - X1, 0 To Y2 - Y1)
'
    On Error GoTo LeggeAreaBMP_ERR
'
    ' N. di pixels che stanno in 3 bytes:
    Pi3B = Switch(BxP = 1, 24, BxP = 4, 6 _
                , BxP = 8, 3, BxP = 24, 1)
    ReDim Col(0 To Pi3B - 1) ' Vettore dei colori dei pixels che stanno in 3 bytes.
'
    ' Calcolo il byte di inizio lettura
    ' all' interno della riga:
    InX0 = CLng(3! * CSng(X1) / CSng(Pi3B))
    ' Calcolo il bit di inizio
    ' all' interno del primo byte:
    If BxP = 1 Then
        InB0 = X1 Mod 8
    ElseIf BxP = 4 Then
        InB0 = X1 Mod 2
    Else
        InB0 = 0
    End If
'
    YIP = Y1
    Do
        InY = BMPHeight - YIP - 1       ' Il BMP lavora dal basso all' alto.
        BnY = InY * XBMP + OffSet       ' Posizione, nel file, del byte corrispondente
                                        ' all' inizio della InYma riga.
        InB = InB0
        InX = InX0
        XIP = X1
        Do
            BnX = BnY + InX     ' Posizione, nel file, del byte corrispondente a InX
                                ' all' interno della InYma riga.
'
            ' Leggo dal file 3 bytes per volta:
            Get FB, BnX, Cl
            Cl = Cl And &HFFFFFF
            InX = InX + 3
'
            ' Leggo il colore dei pixels contenuti nei 3 bytes.
            ' Nel caso di 1, 4 e 8 BitsPerPixel il valore letto
            ' dal file e' un indice nella tabella dei colori TCol():
            If BxP = 1 Then
                ' 24 pixels da leggere:
                C = 23
                For C1 = 2 To 0 Step -1
                    Cl1 = (Cl \ (&H100 ^ C1)) And &HFF
                    For C2 = 0 To 7
                        If (Cl1 And (2 ^ C2)) = 0 Then
                            Col(C) = TCol(0)
                        Else
                            Col(C) = TCol(1)
                        End If
                        C = C - 1
                    Next C2
                Next C1
'
            ElseIf BxP = 4 Then
                ' 6 pixels da leggere:
                Col(4) = TCol(Cl \ &H100000)
                Col(5) = TCol(Cl \ &H10000 And &HF)
                Col(2) = TCol(Cl \ &H1000 And &HF)
                Col(3) = TCol(Cl \ &H100 And &HF)
                Col(0) = TCol(Cl \ &H10 And &HF)
                Col(1) = TCol(Cl And &HF)
'
            ElseIf BxP = 8 Then
                ' 3 pixels da leggere:
                Col(2) = TCol(Cl \ &H10000)
                Col(1) = TCol(Cl \ &H100 And &HFF)
                Col(0) = TCol(Cl And &HFF)
'
            ElseIf BxP = 24 Then
                ' 1 solo pixel da leggere direttamente dal file:
                Col(0) = RGB(Cl \ &H10000, (Cl \ &H100) And &HFF, Cl And &HFF)
            End If
'
            For C = InB To Pi3B - 1
                MP(XIP - X1, YIP - Y1) = Col(C)
                XIP = XIP + 1
                If XIP > X2 Then Exit Do
            Next C
            InB = 0
        Loop While XIP <= X2
'
        DoEvents
        If Termina Then Exit Do
'
        YIP = YIP + 1
    Loop While YIP <= Y2
'
'
LeggeAreaBMP_ERR:
    LeggeAreaBMP = (Err = 0) And (Not Termina)
'
'
'
End Function
Private Sub Swap(N1 As Variant, N2 As Variant)
'
'
    Dim T As Variant
'
    T = N1
    N1 = N2
    N2 = T
'
'
'
End Sub
