Attribute VB_Name = "modSalvaBMP"
'==============================================================
' Descrizione.....: Routines per creare files BMP a profondita'
'                   di colore specificata.
' Nome dei Files..: SalvaBMP.bas, StruttureBMP.bas
' Data............: 4/11/2000
' Versione........: 1.0 a 32 bits
' Sistema.........: Visual Basic 6.0
'                   sotto Windows NT 4.0 (SP 3).
' Scritto da......: F. Languasco 
' E-Mail..........: MC7061@mclink.it
'==============================================================
'
'   Comprende:
'   SalvaImmagineBMP(PBSrc As PictureBox, FileBMP$, iPdC&)
'    PBSrc:     PictureBox di cui si vuole
'               salvare l' immagine.
'    FileBMP$:  Nome del file completo di Path.
'    iPdC:      Indice della profondita' di colore desiderata.
'     1:        TrueColor   24 [BitsPerPixel].
'     2:        256 colori   8 [BitsPerPixel].
'     3:        256 grigi    8 [BitsPerPixel].
'     4:        16 colori    4 [BitsPerPixel].
'     5:        16 grigi     4 [BitsPerPixel].
'     6:        Bianco/Nero  1 [BitsPerPixel].
'     XPPM:     Numero di pixels per metro orizzontale [opzionale),
'     YPPM:     Numero di pixels per metro verticale [opzionale):
'               se non conosciuti possono essere lasciati a 0.
'
'   Nota (forse pleonastica):
'   Le routines di questo modulo sono utili per convertire
'   la profondita' di colore di un' immagine verso valori
'   decrescenti (e.g. TrueColor -> 256 c. -> B/N).
'   La conversione in ordine inverso, benche' possibile, ha
'   poco senso, perche' aumenterebbe solo le dimensioni del
'   file senza aumentare il contenuto di informazione.
'   Eccezione, per i pignoli: un' immagine di pochi pixels
'   ha un file piu' corto se salvata come TrueColor che se
'   salvata a 256 colori! (Usare la Function LunghezzaFileBMP
'   per calcolare il miglior compromesso).
'
Option Explicit
'
Private hMem&, lpBits&, bi As BITMAPINFO_Type   ' Variabili per la
Private BytesNeeded&, BytesPerScanLine&         ' creazione di un
Private bmfHeader As BITMAPFILEHEADER_Type      ' DIB.
'
' Temporary files API:
Private Const MAX_PATH& = 260
Private Declare Function GetTempPath& Lib "kernel32" Alias _
"GetTempPathA" (ByVal nBufferLength As Long, ByVal lpBuffer As String)
Private Declare Function GetTempFileName& Lib "kernel32" Alias _
"GetTempFileNameA" (ByVal lpszPath As String _
, ByVal lpPrefixString As String, ByVal wUnique As Long _
, ByVal lpTempFileName As String)
Private Function FileExists(ByVal N_File$) As Boolean
'
'   Verifica l' esistenza del file N_File$.
'   N_File$ e' il nome completo di path.
'
    Dim FL&
'
    On Error Resume Next
'
    FL = FileLen(N_File$)
    FileExists = (Err = 0)
    Err.Clear
'
'
'
End Function

Private Function MemBuffer(ByVal BxP& _
    , ByVal F_Width&, ByVal F_Height&) As Long
'
'   Ritorna il numero di Bytes necessario a memorizzare una
'   immagine di F_Width x F_Height [Pixels] tenendo conto del
'   numero di colori BxP [BitsPerPixel]. Imposta anche la variabile
'   BytesPerScanLine sulla lunghezza vera, in [Bytes], della riga.
'   Ritorna 0 se il numero di colori e' sconosciuto.
'
    Select Case BxP
        Case 1
        BytesPerScanLine = (F_Width + 7) \ 8
        Case 4
        BytesPerScanLine = (F_Width + 1) \ 2
        Case 8
        BytesPerScanLine = F_Width
        Case 16
        BytesPerScanLine = F_Width * 2
        Case 24
        BytesPerScanLine = F_Width * 3
        Case 32
        BytesPerScanLine = F_Width * 4
        Case Else
        MemBuffer = 0
        Exit Function
    End Select
'
    BytesPerScanLine = ScanAlign(BytesPerScanLine)
    MemBuffer = BytesPerScanLine * F_Height
'
'
'
End Function
Private Function ScanAlign(ByVal Valore&) As Long
'
'   Ritorna un multiplo di 4 uguale o immediatamente
'   superiore a Valore&.
'
    ScanAlign = (Valore + 3&) And &HFFFFFFFC
'
'
'
End Function


Private Function Legge_PictureBox(ByVal PBSrc As PictureBox _
    , ByVal BxP&, ByVal XPPM&, ByVal YPPM&) As Boolean
'
'   Legge l' immagine del PictureBox PBSrc e la memorizza, per usi
'   successivi, nelle variabili hMem&, lpBits& e bi.
'   Il DIB viene creato a BxP [BitsPerPixel].
'
    Dim R1&, hDC_Mem&, M$
    Dim PB_Width&, PB_Height&, Rect As RECT_Type
'
    On Error GoTo Legge_PictureBox_ERR
'
    ' Create a temporary memory DC and select into it
    ' the background picture of the PBSrc Picture Box:
    hDC_Mem = CreateCompatibleDC(PBSrc.hDC)
'
    ' Get the size of the picture bitmap:
    R1 = GetClientRect(PBSrc.hWnd, Rect)
    PB_Width = Abs(Rect.Right - Rect.Left)
    PB_Height = Abs(Rect.Bottom - Rect.Top)
'
    ' Fill the BITMAPINFOHEADER for the desired DIB:
    bi.bmiHeader.biSize = Len(bi.bmiHeader)
    bi.bmiHeader.biWidth = PB_Width
    bi.bmiHeader.biHeight = PB_Height
    bi.bmiHeader.biPlanes = 1           ' ColorPlanes
    bi.bmiHeader.biBitCount = CInt(BxP) ' BitsPerPixel * ColorPlanes
    bi.bmiHeader.biCompression = BI_RGB
    bi.bmiHeader.biSizeImage = 0
    bi.bmiHeader.biXPelsPerMeter = XPPM
    bi.bmiHeader.biYPelsPerMeter = YPPM
    bi.bmiHeader.biClrUsed = 0
    bi.bmiHeader.biClrImportant = 0
'
    ' Calculate the number of bytes needed to store the image:
    BytesNeeded = MemBuffer(bi.bmiHeader.biBitCount _
    , bi.bmiHeader.biWidth, bi.bmiHeader.biHeight)
    If BytesNeeded = 0 Then
        M$ = "Formato grafico sconosciuto." & vbCrLf
        Err.Raise 1001
    End If
'
    ' Allocate a buffer to hold the bitmap bits:
    hMem = GlobalAlloc(GMEM_MOVEABLE, BytesNeeded)
'
    If hMem <> 0 Then
        lpBits = GlobalLock&(hMem)
        ' Get the bits that make up the image:
        R1 = GetDIBits(hDC_Mem, PBSrc.Picture, 0 _
        , PB_Height, ByVal lpBits, bi, DIB_RGB_COLORS)
'
    Else
        M$ = "hMem = " & Str$(hMem) & vbCrLf
        Err.Raise 1001
    End If
'
'
Legge_PictureBox_ERR:
    Legge_PictureBox = (Err = 0)
    If Err <> 0 Then
        M$ = M$ & "Errore N. " & Err.Number & vbCrLf & Err.Description
        MsgBox M$, vbCritical, " Errore in Legge_PictureBox"
    End If
'
    R1 = DeleteDC(hDC_Mem)
'
'
'
End Function



Public Function SalvaImmagineBMP(ByVal PBSrc As PictureBox _
    , ByVal FileBMP$, ByVal iPdC& _
    , Optional ByVal XPPM& = 0 _
    , Optional ByVal YPPM& = 0) As Boolean
'
'   Salva l' immagine di un PictureBox su di un file
'   .BMP con la profondita' di colore specificata.
'
'   Entra con:
'    PBSrc:     PictureBox di cui si vuole
'               salvare l' immagine.
'    FileBMP$:  Nome del file completo di Path.
'    iPdC:      Indice della profondita' di colore desiderata.
'     1:        TrueColor   24 [BitsPerPixel].
'     2:        256 colori   8 [BitsPerPixel].
'     3:        256 grigi    8 [BitsPerPixel].
'     4:        16 colori    4 [BitsPerPixel].
'     5:        16 grigi     4 [BitsPerPixel].
'     6:        Bianco/Nero  1 [BitsPerPixel].
'     XPPM:     Numero di pixels per metro orizzontale [opzionale).
'     YPPM:     Numero di pixels per metro verticale [opzionale).
'   Ritorna True se le operazioni sono state
'   completate con successo.
'
    Dim BxP&, R1&, FileN$, M$
'
    On Error GoTo SalvaImmagineBMP_ERR
'
    FileN$ = TempFile$                      ' Blocca l' immagine di Src
    SavePicture PBSrc.Image, FileN$         ' come Background BitMap.
    PBSrc.Picture = LoadPicture("")         ' Funziona anche con scheda
    PBSrc.Picture = LoadPicture(FileN$)     ' video impostata a profondita'
    If FileExists(FileN$) Then Kill FileN$  ' di colore = 8 [bits per pixel].
'
    ' Definisce il numero di Bits per Pixel:
    BxP = Switch(iPdC = 1, 24 _
               , iPdC = 2, 8 _
               , iPdC = 3, 8 _
               , iPdC = 4, 4 _
               , iPdC = 5, 4 _
               , iPdC = 6, 1)
'
    If Legge_PictureBox(PBSrc, BxP, XPPM, YPPM) Then
        If (iPdC = 3) Or (iPdC = 5) Then ImmGrigia BxP
        If Not SalvaBMP(FileBMP$, BxP) Then Err.Raise 1001
    Else
        Err.Raise 1001
    End If
'
'
SalvaImmagineBMP_ERR:
    ' Libera il blocco di memoria contenente
    ' i dati temporanei del DIB:
    If hMem <> 0 Then
        R1 = GlobalUnlock(hMem)
        hMem = GlobalFree(hMem)
    End If
'
    SalvaImmagineBMP = (Err = 0)
    If Err <> 0 And Err <> 1001 Then
        M$ = "Errore N. " & Err.Number & vbCrLf & Err.Description
        MsgBox M$, vbCritical, " Errore in SalvaImmagineBMP"
    End If
'
'
'
End Function
Private Function TempFile$(Optional ByVal I3C$ = "TMP" _
    , Optional ByRef TempPath As String)
'
'   Ritorna il nome di un File temporaneo [da cancellare
'   esplicitamente dopo l' uso].
'   Con I3C$ si possono specificare fino a tre caratteri
'   iniziali per l' identificazione del file:
'
    Dim R1&, lpBuffer$
    Dim TempFileName As String
'
    R1 = GetTempPath(0, "")
    lpBuffer$ = Space(R1)
    R1 = GetTempPath(R1, lpBuffer$)
    TempPath = Left$(lpBuffer$, R1)
'
    TempFileName = Space(MAX_PATH)
    R1 = GetTempFileName(TempPath, I3C$, 0, TempFileName)
    TempFile$ = Trim$(TempFileName)
'
'
'
End Function



Private Function SalvaBMP(ByVal FileBMP$, ByVal BxP&) As Boolean
'
'   Salva il DIB, definito nelle variabili  hMem&,
'   lpBits& e bi, sul file FileBMP$.
'
    Dim C&, FF%, M$
'
    On Error GoTo SalvaBMP_ERR
'
    If FileExists(FileBMP$) Then Kill FileBMP$
'
    FF = FreeFile
    Open FileBMP$ For Binary As FF
'
    ' Scrive prima il BITMAPINFOHEADER:
    Put #FF, 15, bi.bmiHeader
'
    ' Scrive la tavola dei colori:
    If bi.bmiHeader.biBitCount <= 8 Then
        For C = 0 To (2 ^ BxP) - 1
            Put #FF, , bi.bmiColors(C)
        Next C
    End If
    bmfHeader.bfOffBits = Seek(FF) - 1
'
    ' Scrive la tabella dei dati.
    ' Per Windows NT. Puo' dare problemi
    ' con Windows 98 e immagini molto grandi:
    ReDim InBuf(0 To BytesNeeded - 1) As Byte
    MoveMemory InBuf(0), ByVal lpBits, BytesNeeded
    Put #FF, , InBuf
    ' Per Windows 98:
    'Dim InBuf As Byte, ByteAddr&
    'ByteAddr = lpBits
    'For C = 0 To BytesNeeded - 1
    '    MoveMemory InBuf, ByVal ByteAddr, 1
    '    Put #FF, , InBuf
    '    ByteAddr = ByteAddr + 1
    'Next C
    bmfHeader.bfSize = Seek(FF) - 1
'
    ' Scrive, per ultimo, il BITMAPFILEHEADER
    ' con le informazioni raccolte:
    bmfHeader.bfType = "BM"
    Put #FF, 1, bmfHeader
'
'
SalvaBMP_ERR:
    Close FF
'
    SalvaBMP = (Err = 0)
    If Err <> 0 Then
        M$ = "Errore N. " & Err.Number & vbCrLf & Err.Description
        MsgBox M$, vbCritical, " Errore in SalvaBMP"
    End If
'
'
'
End Function

Private Sub ImmGrigia(ByVal BxP&)
'
'   Converte la tavola dei colori, definita
'   in bmiColors(), in toni di grigio:
'
    Dim C&, R&, G&, B&, Gr&
'
    For C = 0 To (2 ^ BxP) - 1
        R = bi.bmiColors(C).rgbRed
        G = bi.bmiColors(C).rgbGreen
        B = bi.bmiColors(C).rgbBlue
'
        Gr = (R + G + B) / 3
        bi.bmiColors(C).rgbRed = Gr
        bi.bmiColors(C).rgbGreen = Gr
        bi.bmiColors(C).rgbBlue = Gr
    Next C
'
'
'
End Sub

Public Function LunghezzaFileBMP(ByVal BxP& _
    , ByVal F_Width&, ByVal F_Height&) As Long
'
'   Ritorna la lunghezza, in [Bytes], di un file .BMP atto
'   ad ospitare un' immagine di F_Width x F_Height [Pixels]
'   tenendo conto del numero di colori BxP [BitsPerPixel]:
'
    Dim LFB&
'
    ' BITMAPFILEHEADER:
    LFB = 14
'
    ' + BITMAPINFOHEADER:
    LFB = LFB + 40
'
    ' + Tabella dei colori:
    Select Case BxP
        Case 1
        ' Immagine Bianco/Nero:
        LFB = LFB + 8
        Case 4
        ' Immagine 16 colori:
        LFB = LFB + 64
        Case 8
        ' Immagine 256 colori:
        LFB = LFB + 1024
        Case 24
        ' Immagine TrueColor:
        LFB = LFB + 0
    End Select
'
    ' + Tabella dei dati:
    LFB = LFB + MemBuffer(BxP, F_Width, F_Height)
'
'
    LunghezzaFileBMP = LFB  ' [Bytes].
'
'
'
End Function
